import torch
from copy import deepcopy
from GANDLF.metrics import get_metrics
from abc import ABC, abstractmethod
from typing import Union


class AbstractMetricCalculator(ABC):
    def __init__(self, params: dict):
        super().__init__()
        self.params = deepcopy(params)
        self._initialize_metrics_dict()

    def _initialize_metrics_dict(self):
        self.metrics_calculators = get_metrics(self.params)

    def _process_metric_value(self, metric_value: Union[torch.Tensor, float]):
        if isinstance(metric_value, float):
            return metric_value
        if metric_value.dim() == 0:
            return metric_value.item()
        else:
            return metric_value.tolist()

    @staticmethod
    def _inject_kwargs_into_params(params, **kwargs):
        for key, value in kwargs.items():
            params[key] = value
        return params

    @abstractmethod
    def __call__(
        self, prediction: torch.Tensor, target: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        pass


class MetricCalculatorSDNet(AbstractMetricCalculator):
    def __init__(self, params):
        super().__init__(params)

    def __call__(self, prediction: torch.Tensor, target: torch.Tensor, **kwargs):
        params = deepcopy(self.params)
        params = self._inject_kwargs_into_params(params, **kwargs)

        metric_results = {}

        for metric_name, metric_calculator in self.metrics_calculators.items():
            metric_value = (
                metric_calculator(prediction[0], target.squeeze(-1), params)
                .detach()
                .cpu()
            )
            metric_results[metric_name] = self._process_metric_value(metric_value)
        return metric_results


class MetricCalculatorDeepSupervision(AbstractMetricCalculator):
    def __init__(self, params):
        super().__init__(params)

    def __call__(self, prediction: torch.Tensor, target: torch.Tensor, **kwargs):
        params = deepcopy(self.params)
        params = self._inject_kwargs_into_params(params, **kwargs)
        metric_results = {}

        for metric_name, metric_calculator in self.metrics_calculators.items():
            metric_results[metric_name] = 0.0
            for i, _ in enumerate(prediction):
                metric_value = (
                    metric_calculator(prediction[i], target[i], params).detach().cpu()
                )
                metric_results[metric_name] += self._process_metric_value(metric_value)
        return metric_results


class MetricCalculatorSimple(AbstractMetricCalculator):
    def __init__(self, params):
        super().__init__(params)

    def __call__(self, prediction: torch.Tensor, target: torch.Tensor, **kwargs):
        params = deepcopy(self.params)
        params = self._inject_kwargs_into_params(params, **kwargs)
        metric_results = {}

        for metric_name, metric_calculator in self.metrics_calculators.items():
            metric_value = metric_calculator(prediction, target, params).detach().cpu()
            metric_results[metric_name] = self._process_metric_value(metric_value)
        return metric_results


class MetricCalculatorFactory:
    def __init__(self, params: dict):
        self.params = params

    def get_metric_calculator(self) -> AbstractMetricCalculator:
        if self.params["model"]["architecture"] == "sdnet":
            return MetricCalculatorSDNet(self.params)
        elif "deep" in self.params["model"]["architecture"].lower():
            return MetricCalculatorDeepSupervision(self.params)
        else:
            return MetricCalculatorSimple(self.params)



from pathlib import Path
import os

from typing import Optional
import numpy as np

from panoptica import Panoptica_Evaluator


def generate_instance_segmentation(
    prediction: np.ndarray,
    target: np.ndarray,
    parameters: dict,
    panoptica_config_path: Optional[str] = None,
) -> dict:
    """
    Evaluate a single exam using Panoptica.

    Args:
        prediction (np.ndarray): The input prediction containing objects.
        label_path (str): The path to the reference label.
        target (np.ndarray): The input target containing objects.
        parameters (dict): The GaNDLF parameters from which panoptica config is to be extracted.
        panoptica_config_path (str): The path to the Panoptica configuration file.

    Returns:
        dict: The evaluation results.
    """

    os.environ["PANOPTICA_CITATION_REMINDER"] = "False"

    # the parameters dict takes precedence over the panoptica_config_path
    evaluator = parameters.get("panoptica_config", None)
    if evaluator is None:
        cwd = Path(__file__).parent.absolute()
        panoptica_config_path = (
            str(cwd / "panoptica_config_brats.yaml")
            if panoptica_config_path is None
            else panoptica_config_path
        )
        evaluator = Panoptica_Evaluator.load_from_config(panoptica_config_path)

    assert evaluator is not None, "Panoptica evaluator could not be initialized."

    # call evaluate
    group2result = evaluator.evaluate(prediction_arr=prediction, reference_arr=target)

    results = {k: r.to_dict() for k, r in group2result.items()}
    return results



# -*- coding: utf-8 -*-
"""
Implementation of Light UNet
"""
import torch
from torch.nn import ModuleList

from GANDLF.models.seg_modules.DownsamplingModule import DownsamplingModule
from GANDLF.models.seg_modules.EncodingModule import EncodingModule
from GANDLF.models.seg_modules.DecodingModule import DecodingModule
from GANDLF.models.seg_modules.UpsamplingModule import UpsamplingModule
from GANDLF.models.seg_modules.InitialConv import InitialConv
from GANDLF.models.seg_modules.out_conv import out_conv
from .modelBase import ModelBase


class light_unet_multilayer(ModelBase):
    """
    This is the LIGHT U-Net architecture.
    """

    def __init__(self, parameters: dict, residualConnections=False):
        self.network_kwargs = {"res": residualConnections}
        super(light_unet_multilayer, self).__init__(parameters)

        parameters["model"]["depth"] = parameters["model"].get("depth", 4)

        self.depth = self.model_depth_check(parameters)

        self.ins = InitialConv(
            input_channels=self.n_channels,
            output_channels=self.base_filters,
            conv=self.Conv,
            dropout=self.Dropout,
            norm=self.Norm,
            network_kwargs=self.network_kwargs,
        )

        self.ds = ModuleList([])
        self.en = ModuleList([])
        self.us = ModuleList([])
        self.de = ModuleList([])

        for _ in range(0, self.depth):
            self.ds.append(
                DownsamplingModule(
                    input_channels=self.base_filters,
                    output_channels=self.base_filters,
                    conv=self.Conv,
                    norm=self.Norm,
                )
            )

            self.us.append(
                UpsamplingModule(
                    input_channels=self.base_filters,
                    output_channels=self.base_filters,
                    conv=self.Conv,
                    interpolation_mode=self.linear_interpolation_mode,
                )
            )

            self.de.append(
                DecodingModule(
                    input_channels=self.base_filters * 2,
                    output_channels=self.base_filters,
                    conv=self.Conv,
                    norm=self.Norm,
                    network_kwargs=self.network_kwargs,
                )
            )

            self.en.append(
                EncodingModule(
                    input_channels=self.base_filters,
                    output_channels=self.base_filters,
                    conv=self.Conv,
                    dropout=self.Dropout,
                    norm=self.Norm,
                    network_kwargs=self.network_kwargs,
                )
            )

        self.out = out_conv(
            input_channels=self.base_filters,
            output_channels=self.n_classes,
            conv=self.Conv,
            norm=self.Norm,
            network_kwargs=self.network_kwargs,
            final_convolution_layer=self.final_convolution_layer,
            sigmoid_input_multiplier=self.sigmoid_input_multiplier,
        )

        if "converter_type" in parameters["model"]:
            self.ins = self.converter(self.ins).model
            self.out = self.converter(self.out).model
            for i_lay in range(0, self.depth):
                self.ds[i_lay] = self.converter(self.ds[i_lay]).model
                self.us[i_lay] = self.converter(self.us[i_lay]).model
                self.de[i_lay] = self.converter(self.de[i_lay]).model
                self.en[i_lay] = self.converter(self.en[i_lay]).model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the network.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        y = []
        y.append(self.ins(x))

        # [downsample --> encode] x num layers
        for i in range(0, self.depth):
            temp = self.ds[i](y[i])
            y.append(self.en[i](temp))

        x = y[-1]

        # [upsample --> encode] x num layers
        for i in range(self.depth - 1, -1, -1):
            x = self.us[i](x)
            x = self.de[i](x, y[i])

        x = self.out(x)
        return x


class light_resunet_multilayer(light_unet_multilayer):
    """
    This is the LIGHT U-Net architecture with residual connections.
    """

    def __init__(self, parameters: dict):
        super(light_resunet_multilayer, self).__init__(
            parameters, residualConnections=True
        )



from torch.optim.lr_scheduler import (
    LambdaLR,
    CyclicLR,
    ExponentialLR,
    StepLR,
    ReduceLROnPlateau,
    CosineAnnealingWarmRestarts,
    CosineAnnealingLR,
)
import math


def cyclical_lr(stepsize, min_lr, max_lr):
    # Scaler : we can adapt this if we do not want the triangular LR
    scaler = lambda x: 1
    # Lambda function to calculate the LR
    lr_lambda = lambda it: max_lr - (max_lr - min_lr) * relative(it, stepsize)

    # Additional function to see where on the cycle we are
    def relative(it, stepsize):
        cycle = math.floor(1 + it / (2 * stepsize))
        x = abs(it / stepsize - 2 * cycle + 1)
        return max(0, (1 - x)) * scaler(cycle)

    return lr_lambda


def cyclical_lr_modified(cycle_length, min_lr, max_lr, max_lr_multiplier):
    min_lr_multiplier = min_lr / max_lr
    max_lr_multiplier = 1.0
    # Lambda function to calculate what to multiply the initial learning rate by
    # The beginning and end of the cycle result in highest multipliers (lowest at the center)
    mult = lambda it: max_lr_multiplier * rel_dist(
        it, cycle_length
    ) + min_lr_multiplier * (1 - rel_dist(it, cycle_length))

    def rel_dist(iteration, cycle_length):
        # relative_distance from iteration to the center of the cycle
        # equal to 1 at beggining of cycle and 0 right at the cycle center

        # reduce the iteration to less than the cycle length
        iteration = iteration % cycle_length
        return 2 * abs(iteration - cycle_length / 2.0) / cycle_length

    return mult


def base_triangle(parameters):
    """
    This function parses the parameters from the config file and returns the appropriate object
    """

    # pick defaults
    parameters["scheduler"]["min_lr"] = parameters["scheduler"].get("min_lr", 10**-3)
    parameters["scheduler"]["max_lr"] = parameters["scheduler"].get("max_lr", 1)

    clr = cyclical_lr(
        parameters["scheduler"]["step_size"],
        min_lr=parameters["scheduler"]["min_lr"],
        max_lr=parameters["scheduler"]["max_lr"],
    )
    return LambdaLR(parameters["optimizer_object"], [clr])


def triangle_modified(parameters):
    # pick defaults
    parameters["scheduler"]["min_lr"] = parameters["scheduler"].get("min_lr", 0.000001)
    parameters["scheduler"]["max_lr"] = parameters["scheduler"].get("max_lr", 0.001)
    parameters["scheduler"]["max_lr_multiplier"] = parameters["scheduler"].get(
        "max_lr_multiplier", 1.0
    )

    clr = cyclical_lr_modified(
        parameters["scheduler"]["step_size"],
        parameters["scheduler"]["min_lr"],
        parameters["scheduler"]["max_lr"],
        parameters["scheduler"]["max_lr_multiplier"],
    )
    return LambdaLR(parameters["optimizer_object"], [clr])


def cyclic_lr_base(parameters, mode="triangular"):
    # pick defaults for "min_lr", "max_lr", "max_lr_multiplier" if not present in parameters
    parameters["scheduler"]["min_lr"] = parameters["scheduler"].get(
        "min_lr", parameters["learning_rate"] * 0.001
    )
    parameters["scheduler"]["max_lr"] = parameters["scheduler"].get(
        "max_lr", parameters["learning_rate"]
    )
    parameters["scheduler"]["gamma"] = parameters["scheduler"].get("gamma", 0.1)
    parameters["scheduler"]["scale_mode"] = parameters["scheduler"].get(
        "scale_mode", "cycle"
    )
    parameters["scheduler"]["cycle_momentum"] = parameters["scheduler"].get(
        "cycle_momentum", False
    )
    parameters["scheduler"]["base_momentum"] = parameters["scheduler"].get(
        "base_momentum", 0.8
    )
    parameters["scheduler"]["max_momentum"] = parameters["scheduler"].get(
        "max_momentum", 0.9
    )

    return CyclicLR(
        parameters["optimizer_object"],
        parameters["learning_rate"] * 0.001,  # min lr
        parameters["learning_rate"],  # mar_lr
        step_size_up=parameters["scheduler"]["step_size"],
        step_size_down=None,
        mode=mode,
        gamma=parameters["scheduler"]["gamma"],
        scale_fn=None,
        scale_mode=parameters["scheduler"]["scale_mode"],
        cycle_momentum=parameters["scheduler"]["cycle_momentum"],
        base_momentum=parameters["scheduler"]["base_momentum"],
        max_momentum=parameters["scheduler"]["max_momentum"],
    )


## this is not working with default step_size, for some reason
# def cyclic_lr_triangular2(parameters):
#     return cyclic_lr_base(parameters, mode="triangular2")


def cyclic_lr_exp_range(parameters):
    return cyclic_lr_base(parameters, mode="exp_range")


def exp(parameters):
    parameters["scheduler"]["gamma"] = parameters["scheduler"].get("gamma", 0.1)
    return ExponentialLR(
        parameters["optimizer_object"], parameters["scheduler"]["gamma"]
    )


def step(parameters):
    parameters["scheduler"]["gamma"] = parameters["scheduler"].get("gamma", 0.1)
    return StepLR(
        parameters["optimizer_object"],
        parameters["scheduler"]["step_size"],
        gamma=parameters["scheduler"]["gamma"],
    )


def reduce_on_plateau(parameters):
    parameters["scheduler"]["min_lr"] = parameters["scheduler"].get(
        "min_lr", parameters["learning_rate"] * 0.001
    )
    parameters["scheduler"]["gamma"] = parameters["scheduler"].get("gamma", 0.1)
    parameters["scheduler"]["mode"] = parameters["scheduler"].get("mde", "min")
    parameters["scheduler"]["threshold_mode"] = parameters["scheduler"].get(
        "threshold_mode", "rel"
    )
    parameters["scheduler"]["factor"] = parameters["scheduler"].get("factor", 0.1)
    parameters["scheduler"]["patience"] = parameters["scheduler"].get("patience", 10)
    parameters["scheduler"]["threshold"] = parameters["scheduler"].get(
        "threshold", 0.0001
    )
    parameters["scheduler"]["cooldown"] = parameters["scheduler"].get("cooldown", 0)

    return ReduceLROnPlateau(
        parameters["optimizer_object"],
        mode=parameters["scheduler"]["mode"],
        factor=parameters["scheduler"]["factor"],
        patience=parameters["scheduler"]["patience"],
        threshold=parameters["scheduler"]["threshold"],
        threshold_mode=parameters["scheduler"]["threshold_mode"],
        cooldown=parameters["scheduler"]["cooldown"],
        min_lr=parameters["scheduler"]["min_lr"],
    )


def cosineannealingwarmrestarts(parameters):
    parameters["scheduler"]["T_0"] = parameters["scheduler"].get("T_0", 5)
    parameters["scheduler"]["T_mult"] = parameters["scheduler"].get("T_mult", 1)
    parameters["scheduler"]["eta_min"] = parameters["scheduler"].get("eta_min", 0.001)

    return CosineAnnealingWarmRestarts(
        parameters["optimizer_object"],
        T_0=parameters["scheduler"]["T_0"],
        T_mult=parameters["scheduler"]["T_mult"],
        eta_min=parameters["scheduler"]["eta_min"],
    )


def cosineannealingLR(parameters):
    parameters["scheduler"]["T_max"] = parameters["scheduler"].get("T_max", 50)
    parameters["scheduler"]["eta_min"] = parameters["scheduler"].get("eta_min", 0.001)

    return CosineAnnealingLR(
        parameters["optimizer_object"],
        T_max=parameters["scheduler"]["T_max"],
        eta_min=parameters["scheduler"]["eta_min"],
    )



import os
import yaml

# codacy ignore python-use-of-pickle: Pickle usage is safe in this context (local data only).
import pickle
import shutil
import pandas as pd
from pathlib import Path
from warnings import warn

import lightning.pytorch as pl
from lightning.pytorch.profilers import PyTorchProfiler
from lightning.pytorch.tuner import Tuner as LightningTuner

from GANDLF.utils import get_dataframe, split_data
from GANDLF.models.lightning_module import GandlfLightningModule
from GANDLF.data.lightning_datamodule import GandlfTrainingDatamodule

from typing import Optional


def TrainingManager(
    dataframe: pd.DataFrame,
    outputDir: str,
    parameters: dict,
    resume: bool,
    reset: bool,
    profile: Optional[bool] = False,
) -> None:
    """
    This is the training manager that ties all the training functionality together

    Args:
        dataframe (pandas.DataFrame): The full data from CSV.
        outputDir (str): The main output directory.
        parameters (dict): The parameters dictionary.
        resume (bool): Whether the previous run will be resumed or not.
        reset (bool): Whether the previous run will be reset or not.
        profile(bool): Whether we want the profile activity or not. Defaults to False.

    """

    if "output_dir" not in parameters:
        parameters["output_dir"] = outputDir
    if reset:
        shutil.rmtree(outputDir)
        Path(outputDir).mkdir(parents=True, exist_ok=True)

    # save the current model configuration as a sanity check
    currentModelConfigPickle = os.path.join(outputDir, "parameters.pkl")
    if (not os.path.exists(currentModelConfigPickle)) or reset or resume:
        with open(currentModelConfigPickle, "wb") as handle:
            pickle.dump(parameters, handle, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        if os.path.exists(currentModelConfigPickle):
            print(
                "Using previously saved parameter file",
                currentModelConfigPickle,
                flush=True,
            )
            parameters = pickle.load(open(currentModelConfigPickle, "rb"))

    dataframe_split = split_data(dataframe, parameters)

    last_indeces, _, _, _ = dataframe_split[-1]

    # check the last indeces to see if single fold training is requested
    singleFoldTesting = True if last_indeces[0] == 0 else False
    singleFoldValidation = True if last_indeces[1] == 0 else False

    for (
        testing_and_valid_indeces,
        trainingData,
        validationData,
        testingData,
    ) in dataframe_split:
        # the output of the current fold is only needed if multi-fold training is happening
        currentTestingOutputFolder = outputDir
        if not singleFoldTesting:
            currentTestingOutputFolder = os.path.join(
                outputDir, "testing_" + str(testing_and_valid_indeces[0])
            )
            Path(currentTestingOutputFolder).mkdir(parents=True, exist_ok=True)

        currentValidationOutputFolder = currentTestingOutputFolder
        if not singleFoldValidation:
            currentValidationOutputFolder = os.path.join(
                currentTestingOutputFolder, str(testing_and_valid_indeces[1])
            )
            Path(currentValidationOutputFolder).mkdir(parents=True, exist_ok=True)

        # initialize the dataframes and save them to disk
        data_dict = {
            "training": trainingData,
            "validation": validationData,
            "testing": testingData,
        }
        data_dict_files = {}
        for data_type, data in data_dict.items():
            data_dict_files[data_type] = None
            if data is not None:
                currentDataPickle = os.path.join(
                    currentValidationOutputFolder, "data_" + data_type + ".pkl"
                )
                data_dict_files[data_type] = currentDataPickle
                if (not os.path.exists(currentDataPickle)) or reset or resume:
                    data.to_pickle(currentDataPickle)
                    data.to_csv(currentDataPickle.replace(".pkl", ".csv"), index=False)
                else:
                    # read the data from the pickle if present
                    data_dict[data_type] = get_dataframe(currentDataPickle)

        # Dataloader initialization - should be extracted somewhere else (preferably abstracted away)
        datamodule = GandlfTrainingDatamodule(data_dict_files, parameters)
        parameters = datamodule.updated_parameters_dict

        # This entire section should be handled in config parser

        accelerator = parameters.get("accelerator", "auto")
        allowed_accelerators = ["cpu", "gpu", "auto"]
        # codacy ignore Generic/ReDoS: This is not a SQL query, it's an error message.
        assert (
            accelerator in allowed_accelerators
        ), f"Invalid accelerator selected: {accelerator}. Please select from {allowed_accelerators}"
        strategy = parameters.get("strategy", "auto")
        allowed_strategies = ["auto", "ddp"]
        # codacy ignore Generic/ReDoS: This is not a SQL query, it's an error message.
        assert (
            strategy in allowed_strategies
        ), f"Invalid strategy selected: {strategy}. Please select from {allowed_strategies}"
        precision = parameters.get("precision", "32")
        allowed_precisions = [
            "64",
            "64-true",
            "32",
            "32-true",
            "16",
            "16-mixed",
            "bf16",
            "bf16-mixed",
        ]
        # codacy ignore Generic/ReDoS: This is not a SQL query, it's an error message.
        assert (
            precision in allowed_precisions
        ), f"Invalid precision selected: {precision}. Please select from {allowed_precisions}"

        warn(
            f"Configured to use {accelerator} with {strategy} for training, but current development configuration will force single-device only training."
        )
        trainer = pl.Trainer(
            accelerator=accelerator,
            strategy=strategy,
            fast_dev_run=False,
            devices=parameters.get("devices", "auto"),
            num_nodes=parameters.get("num_nodes", 1),
            precision=precision,
            gradient_clip_algorithm=parameters["clip_mode"],
            gradient_clip_val=parameters["clip_grad"],
            max_epochs=parameters["num_epochs"],
            sync_batchnorm=False,
            enable_checkpointing=False,
            logger=False,
            num_sanity_val_steps=0,
            profiler=PyTorchProfiler(sort_by="cpu_time_total", row_limit=10)
            if profile
            else None,
        )

        lightning_module = GandlfLightningModule(
            parameters, output_dir=currentValidationOutputFolder
        )

        if parameters.get("auto_batch_size_find", False):
            LightningTuner(trainer).scale_batch_size(
                lightning_module, datamodule=datamodule
            )

        if parameters.get("auto_lr_find", False):
            LightningTuner(trainer).lr_find(lightning_module, datamodule=datamodule)

        trainer.fit(lightning_module, datamodule=datamodule)

        testing_data = data_dict_files.get("testing", None)
        if testing_data:
            trainer.test(lightning_module, datamodule=datamodule)


def TrainingManager_split(
    dataframe_train: pd.DataFrame,
    dataframe_validation: pd.DataFrame,
    dataframe_testing: pd.DataFrame,
    outputDir: str,
    parameters: dict,
    resume: bool,
    reset: bool,
    profile: Optional[bool] = False,
):
    """
    This is the training manager that ties all the training functionality together

    Args:
        dataframe_train (pd.DataFrame): The training data from CSV.
        dataframe_validation (pd.DataFrame): The validation data from CSV.
        dataframe_testing (pd.DataFrame): The testing data from CSV.
        outputDir (str): The main output directory.
        parameters (dict): The parameters dictionary.
        resume (bool): Whether the previous run will be resumed or not.
        reset (bool): Whether the previous run will be reset or not.
        profile(bool): Whether the we want the profile activity or not. Defaults to False.

    """
    currentModelConfigPickle = os.path.join(outputDir, "parameters.pkl")
    currentModelConfigYaml = os.path.join(outputDir, "config.yaml")

    if (not os.path.exists(currentModelConfigPickle)) or reset or resume:
        with open(currentModelConfigPickle, "wb") as handle:
            pickle.dump(parameters, handle, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        if os.path.exists(currentModelConfigPickle):
            print(
                "Using previously saved parameter file",
                currentModelConfigPickle,
                flush=True,
            )
            parameters = pickle.load(open(currentModelConfigPickle, "rb"))

    if (not os.path.exists(currentModelConfigYaml)) or reset or resume:
        with open(currentModelConfigYaml, "w") as handle:
            yaml.dump(parameters, handle, default_flow_style=False)

    data_dict_files = {
        "training": dataframe_train,
        "validation": dataframe_validation,
        "testing": dataframe_testing,
    }

    datamodule = GandlfTrainingDatamodule(data_dict_files, parameters)
    parameters = datamodule.updated_parameters_dict

    # This entire section should be handled in config parser

    accelerator = parameters.get("accelerator", "auto")
    allowed_accelerators = ["cpu", "gpu", "auto"]
    assert (
        accelerator in allowed_accelerators
    ), f"Invalid accelerator selected: {accelerator}. Please select from {allowed_accelerators}"
    strategy = parameters.get("strategy", "auto")
    allowed_strategies = ["auto", "ddp"]
    assert (
        strategy in allowed_strategies
    ), f"Invalid strategy selected: {strategy}. Please select from {allowed_strategies}"
    precision = parameters.get("precision", "32")
    allowed_precisions = [
        "64",
        "64-true",
        "32",
        "32-true",
        "16",
        "16-mixed",
        "bf16",
        "bf16-mixed",
    ]
    assert (
        precision in allowed_precisions
    ), f"Invalid precision selected: {precision}. Please select from {allowed_precisions}"

    trainer = pl.Trainer(
        accelerator=accelerator,
        strategy=strategy,
        fast_dev_run=False,
        devices=parameters.get("devices", "auto"),
        num_nodes=parameters.get("num_nodes", 1),
        precision=precision,
        gradient_clip_algorithm=parameters["clip_mode"],
        gradient_clip_val=parameters["clip_grad"],
        max_epochs=parameters["num_epochs"],
        sync_batchnorm=False,
        enable_checkpointing=False,
        logger=False,
        num_sanity_val_steps=0,
        profiler=PyTorchProfiler(sort_by="cpu_time_total", row_limit=10)
        if profile
        else None,
    )
    lightning_module = GandlfLightningModule(parameters, output_dir=outputDir)

    if parameters.get("auto_batch_size_find", False):
        LightningTuner(trainer).scale_batch_size(
            lightning_module, datamodule=datamodule
        )

    if parameters.get("auto_lr_find", False):
        LightningTuner(trainer).lr_find(lightning_module, datamodule=datamodule)

    trainer.fit(lightning_module, datamodule=datamodule)

    if dataframe_testing is not None:
        trainer.test(lightning_module, datamodule=datamodule)



import torch.nn as nn
from GANDLF.models.seg_modules.Interpolate import Interpolate


class FCNUpsamplingModule(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv=nn.Conv2d,
        conv_kwargs=None,
        scale_factor=2,
        interpolation_mode="trilinear",
    ):
        """
        Upsampling module for the Fully Convolutional Network.

        Args:
            input_channels (int): Number of input channels.
            output_channels (int): Number of output channels.
            conv (nn.Module): Convolution layer.
            conv_kwargs (dict): Convolution layer keyword arguments.
            scale_factor (int): Scale factor for interpolation.
            interpolation_mode (str): Interpolation mode for upsampling.

        """
        nn.Module.__init__(self)
        if conv_kwargs is None:
            conv_kwargs = {"kernel_size": 1, "stride": 1, "padding": 0, "bias": True}

        self.interpolate_kwargs = {
            "size": None,
            "scale_factor": 2 ** (scale_factor - 1),
            "mode": interpolation_mode,
            "align_corners": True,
        }
        self.interpolate = Interpolate(interp_kwargs=self.interpolate_kwargs)

        self.conv0 = conv(input_channels, output_channels, **conv_kwargs)

    def forward(self, x):
        """
        Upsampling the input tensor using the convolution layer and interpolation.

        Args:
            x (torch.Tensor): Input tensor of shape [batch_size, channels, height, width]

        Returns:
            x (torch.Tensor): Returns the upsampled tensor of shape [batch_size, output_channels, height*2, width*2].

        """
        x = self.interpolate(self.conv0(x))
        return x



#!usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from typing import Optional

from deprecated import deprecated
import click

from GANDLF.cli import copyrightMessage, post_training_model_optimization
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import logger_setup


def _optimize_model(
    model: str, config: Optional[str], output_path: Optional[str] = None
):
    if post_training_model_optimization(
        model_path=model, config_path=config, output_path=output_path
    ):
        print("Post-training model optimization successful.")
    else:
        print("Post-training model optimization failed.")


@click.command()
@click.option(
    "--model",
    "-m",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="Path to the model file (ending in '.pth.tar') you wish to optimize.",
)
@click.option(
    "--output-path",
    "-o",
    type=click.Path(file_okay=False, dir_okay=True),
    required=False,
    help="Location to save the optimized model, defaults to location of `model`",
)
@click.option(
    "--config",
    "-c",
    help="The configuration file (contains all the information related to the training/inference session)."
    "Arg value is used if no config in model is found.",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(
    model: str,
    log_file: str,
    config: Optional[str] = None,
    output_path: Optional[str] = None,
):
    """Generate optimized versions of trained GaNDLF models."""

    logger_setup(log_file)
    _optimize_model(model=model, config=config, output_path=output_path)


# old-fashioned way of running gandlf via `gandlf_optimizeModel`.
@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf optimize-model` cli command "
    + "instead of `gandlf_optimizeModel`.\n"
    + "`gandlf_optimizeModel` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF_OptimizeModel",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate optimized versions of trained GaNDLF models.\n\n"
        + copyrightMessage,
    )

    parser.add_argument(
        "-m",
        "--model",
        metavar="",
        type=str,
        help="Path to the model file (ending in '.pth.tar') you wish to optimize.",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--outputdir",
        "--output_path",
        metavar="",
        type=str,
        default=None,
        help="Location to save the optimized model, defaults to location of `model`",
        required=False,
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="",
        type=str,
        default=None,
        required=False,
        help="The configuration file (contains all the information related to the training/inference session). "
        "Arg value is used if no config in model is found.",
    )

    args = parser.parse_args()
    _optimize_model(model=args.model, config=args.config, output_path=args.outputdir)


if __name__ == "__main__":
    old_way()
